#include <stdlib.h>
#include <stdio.h>
#include "set.h"

/* Question 5 */
struct set_base {
  elem e;
  set next;
};

/* Question 6 */
set empty_set() {
  return NULL;
}

/* Auxiliary function */
/*@ requires nothing 
  assigns nothing 
  ensures returns the adresse of a newly allocated bucket containing e
  and pointing toward n */
set cons(elem e, set n) {
  set r = malloc(sizeof (struct set_base));
  if (r == NULL) {
    printf("Cannot allocate memory for cons.\n");
    exit(3);
  }
  r->e = e;
  r->next = n;
  return r;
}

/* Question 7 */
void add_to_set_v1(elem e, set *s) {
  set visit = *s;
  /*@ loop invariant : e is not present between *s (included) 
    and visit (excluded) */
  while (visit != NULL)
    if (visit->e == e)
      return;
    else
      visit = visit->next;
  *s = cons(e, *s);
}
  
/* Question 7, alternative avec ajout à la fin */
void add_to_set(elem e, set *s) {
  set last;
  if (*s == NULL) *s = cons(e, NULL);
  else {
    last = *s;
    /*@ loop invariant e is not present between *s (included) and 
        last (excluded) */
    while (last->next != NULL) 
      if (last->e == e) return;
      else last = last->next;
    if (last->e == e) return;
    last->next = cons(e, NULL);
  }
} 

/* Question 8 */
void print_set(set s) {
  if (s == NULL) printf("{ }");
  else {
    printf("{ ");
    while (s->next != NULL) {
      printf("%c, ", s->e);
      s = s->next;
    }
    printf("%c }", s->e);
  }
}

/* Question 9 */
void remove_from_set(elem e, set *s) {
  set to_be_freed, visit;
  if (*s == NULL) return;
  if ((*s)->e == e) { /* e is in the head of *s */
    to_be_freed = *s;
    *s = (*s)->next;
    free(to_be_freed);
  }
  else {
    visit = *s;
    /*@ loop invariant : e is not present between *s and visit 
        (both included) */
    while (visit->next != NULL && visit->next->e != e)
      visit = visit->next;
    if (visit->next == NULL) return; /* e is not in *s */
    /* otherwise, e is in visit->next 
       we remove this bucket from the list */
    to_be_freed = visit->next; 
    visit->next = visit->next->next;
    free(to_be_freed);
  }
}


/* Question 10 */
set set_union_v1(set a, set b) {
  while (a != NULL) {
    add_to_set(a->e, &b);
    a = a->next;
  }
  return b;
}

/* Question 11 */
set set_difference(set a, set b) {
  while (b != NULL) {
    remove_from_set(b->e, &a);
    b = b->next;
  }
  return a;
}


/* Question 12 */
/*@ requires a and b well-formed lists
  assigns nothing
  ensures return 1 if a is a subset of b, 0 otherwise */
int set_include_v1(set a, set b) {
  set visit_b;
  int is_in_b;
  /* loop invariant all elements between the initial value of a (included)
     and the current value of a (excluded) are present in b */
  while (a != NULL) {
    visit_b = b;
    is_in_b = 0;
    /*@ loop invariant a->e is not between b (included) and visit_b 
        (excluded) */
    while (!is_in_b && visit_b != NULL) 
      if (a->e == visit_b->e)
	is_in_b = 1;
      else
	visit_b = visit_b->next;
    if (!is_in_b) return 0;
    a = a->next;
  }
  return 1;
}

/* Question 12, alternative using an auxiliary function */
/*@ requires s well-formed list
  assigns nothing
  ensures returns 1 if e belongs to s, 0 otherwise */
int mem(elem e, set s) {
  /*@ loop invariant e is not present in all buckets between the
    initial value of s (included) and the current value of s
    (excluded) */
  while (s != NULL)
    if (s->e == e) return 1;
    else s = s->next;
  return 0;
}

int set_include(set a, set b) {
  /* loop invariant all elements between the initial value of a (included)
     and the current value of a (excluded) are present in b */
  while (a != NULL)
    if (mem(a->e, b)) a = a->next;
    else return 0;
  return 1;
}

/* Question 13 */
int set_equal(set a, set b) {
  return set_include(a, b) && set_include(b,a);
}

/* Question 14 */
/*@ requires b and s are valid addresses, *s is a well-formed list 
  assigns *b, *s
  ensures add the bucket *b is the set *s, free b if b->e is already 
  in *s */
void add_bucket_v1(struct set_base *b, set *s) {
  set visit = *s;
  /*@ loop invariant : b->e is not present between *s (included) 
    and visit (excluded) */
  while (visit != NULL)
    if (visit->e == b->e) {
      free(b);
      return;
    }
    else
      visit = visit->next;
  b->next = *s;
  *s = b;
}

/* Question 14, alternative avec ajout à la fin */
void add_bucket(struct set_base *b, set *s) {
  set last;
  if (*s == NULL) { *s = b; b->next = NULL; }
  else {
    last = *s;
    /*@ loop invariant b->e is not present between *s (included) and 
        last (excluded) */
    while (last->next != NULL) 
      if (last->e == b->e) {
	free(b);
	return;
      }
      else last = last->next;
    if (last->e == b->e) {
      free(b);
      return;
    }
    last->next = b;
    b->next = NULL;
  }
}
  

/* Question 15 */
set set_union(set a, set b) {
  set next; 
  while (a != NULL) {
    /* a->next needs to be kept because it is overwritten by add_bucket */
    next = a->next;
    add_bucket(a, &b);
    a = next;
  }
  return b;
}

    
    
